#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2010 British Broadcasting Corporation and Kamaelia Contributors(1)
#
# (1) Kamaelia Contributors are listed in the AUTHORS file and at
#     http://www.kamaelia.org/AUTHORS - please extend this file,
#     not this notice.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------
#

"""\
===================
IRC Channel Logger
===================
Logger writes all traffic it receives to text files, changing files once per
day. It is built using IRC_Client as its core.  



Example Usage
-------------
To log the channel #sillyputty on server my.server.org::

    Logger('#sillyputty', host="my.server.org").run()

It will now log all messages to #kamtest except those prefixed by "[off]".



More Detail
-----------
BasicLogger is a higher-level IRC client that is meant to link to the base
client found in IRCClient.py. It sends command tuples to its "irc" outbox, and
receives them via its "inbox", allowing it to implement login, and ping
response. It uses IRC_Client's tuple-based output format to achieve some
demultiplexing of IRC output as well, though not of the multiple-channel sort.

BasicLogger depends heavily on the LoggerFunctions module. See LoggerFunctions
for a list of queries it responds to, how it formats the date and time, and how
it determines file names. 

Logger ultimately links BasicLogger's "irc" outbox to IRC_Client's "talk" inbox.
It also utilizes two Carousels and SimpleFileWriters. 



How it works
-------------
Logger writes everything it hears to two files in the specified directory. The
filenames are in the format "givenchannelnameDD-MM-YYYY.log" and
"givenchannelnameDD-MM-YYYY.info".

BasicLogger writes all channel output to its "outbox" and all other messages to
its "system" box. Once per loop, it checks the current date against its stored
date. If the date has changed, then it changes the name of its logfiles to
reflect the new date and sends the new names to "log_next" and "info_next".
Logger uses this in conjunction with a Carousel to create a new logfile and
close the old one.

By default BasicLogger uses ::outformat::, defined in IRCClient, to format
messages from IRCClient.SimpleIRCClientPrefab before writing to the log. To
format messages differently, pass in a different function to its "formatter"
keyword. 

Logger simply links BasicLogger with a IRCClient.SimpleIRCClientPrefab and two
Carousel-encapsulated SimpleFileWriters. It also slaps timestamps on messages.
It takes any keyword that BasicLogger or IRCClient.SimpleIRCClientPrefab will
take.



Command Line Usage
------------------
One can run Logger from the command line by entering::

    ./Logger.py \#somechannel desirednickname
"""
from Kamaelia.File.Writing import SimpleFileWriter
from Kamaelia.Chassis.Graphline import Graphline
from Kamaelia.Chassis.Carousel import Carousel
from Axon.Component import component
import Kamaelia.Protocol.IRC.IRCClient
import Kamaelia.Support.Protocol.IRC

import Kamaelia.Apps.IRCLogger.Support
import time, os
import random
import copy



class BasicLogger(component):
    """\
    BasicLogger(channel, **kwargs) -> new BasicLogger component

    Keyword arguments:

    - formatter -- function that takes an output tuple of IRC_Client's and
                   outputs a string. Default outformat from IRCClient.py
    - name      -- nickname of the logger bot. Default "jinnaslogbot"
    - logdir    -- directory logs are to be put into. Default is the directory
                   this module is in.
    """
    Outboxes = {"irc" : "to IRC, for user responses and login",
                "outbox" : "What we're interested in, the traffic over the channel",
                "system" : "Messages directed toward the client, numeric replies, etc.",
                "signal" : "Shutdown handling in the future",

                "log_next" : "for the Log Carousel",
                "info_next" : "for the Info Carousel"
                }

    def __init__(self,
                 channel,
                 formatter=Kamaelia.Support.Protocol.IRC.outformat,
                 name="jinnaslogbot",
                 logdir="",
                 password=None):
        """x.__init__(...) initializes x; see x.__class__.__doc__ for signature"""
        super(BasicLogger, self).__init__()
        self.channel = channel
        self.format = formatter 
        self.name = name
        self.logdir = logdir.rstrip('/') or os.getcwd()
        self.logdir = self.logdir + '/'
        self.logname = ""
        self.infoname = ""
        self.password = password
        self.debugger.addDebugSection("Logger.main", 0)

    def login(self):
        """registers with the IRC server"""
        self.send(("NICK", self.name), "irc")
        self.send(("USER", self.name, self.name, self.name, self.name), "irc")
        if self.password:
            self.send(("PRIVMSG", 'nickserv', "identify " + self.password), "irc")
        self.send(("JOIN", self.channel), "irc")
        
    def main(self):
        """Main loop"""
        self.login()
        self.changeDate()
        yield 1
        
        while True:
	    time.sleep(0.1) # better way to do this?
            if self.currentDateString() != self.lastdatestring:
                self.changeDate()
                
            yield 1 
            while self.dataReady("inbox"):
                data = self.recv("inbox")
                formatted_data = self.format(data)
                if (data[2] == self.channel or data[0] == 'NICK') and formatted_data: #format might return None
                    self.send(formatted_data, "outbox")
                    self.respondToQueries(data) 
                elif formatted_data:
                    self.send(formatted_data, "system")
                    self.respondToPings(data) 

    def respondToPings(self, msg):
        if msg[0] == 'PING':
            self.send(('PONG', msg[1]), 'irc')
            self.send("Sent PONG to %s \n" % msg[1], "system")

    def respondToQueries(self, msg):
        if msg[0] == 'PRIVMSG' and msg[3].split(':')[0] == self.name:
            words = msg[3].split()[1:]
            if len(words) > 1 and words[0] == "reload":
                try:
                    exec("reload(%s)" % words[1])
                    reply = "'%s' reloaded\n" % words[1]
                except:
                    reply = "'%s' isn't a module, or at least not one I can reload.\n" % words[1]
                self.send(('PRIVMSG', self.channel, reply), "irc")
                self.send(self.format(reply), "outbox")
        Kamaelia.Apps.IRCLogger.Support.respondToQueries(self, msg)

    def currentDateString(self):
        """returns the current date"""
        return Kamaelia.Apps.IRCLogger.Support.currentDateString()

    def currentTimeString(self):
        """returns current time"""
        return Kamaelia.Apps.IRCLogger.Support.currentTimeString()

    def getFilenames(self):
        """returns tuple (logname, infoname) according to the parameters given"""
        return Kamaelia.Apps.IRCLogger.Support.getFilenames(self.logdir, self.channel)
    
    def changeDate(self):
        """updates the date and requests new log files to reflect the date"""
        self.lastdatestring = self.currentDateString()
        self.logname, self.infoname = self.getFilenames()
        self.send(self.logname, "log_next")
        self.send(self.infoname, "info_next")

    
def Logger(channel,
           name=None,
           formatter=Kamaelia.Apps.IRCLogger.Support.TimedOutformat,
           logdir="",
           password=None,
           filewriter = Kamaelia.Apps.IRCLogger.Support.AppendingFileWriter,
           **irc_args):
    """\
    Logger(channel, **kwargs) ->
        Prefab that links the IRC components to BasicLogger
        and two Carousel-encapsulated AppendingFileWriters

    Keyword arguments:

    - formatter -- formatter to run incoming messages from IRC_Client through
      before writing to the log. Default TimedOutformat.
    - name      -- nickname of the loggerbot. Default is the default name defined in
                   BasicLogger.
    - logdir    -- directory logs are to be put into. Default is the directory
                   this module is in.
    - **irc_args  -- pointer to a dictionary containing arguments for IRCClient.SimpleIRCClientPrefab
    """
    return Graphline(irc = Kamaelia.Protocol.IRC.IRCClient.SimpleIRCClientPrefab(**irc_args),
                     logger = BasicLogger(channel, name=name, formatter=formatter, logdir=logdir, password=password),
                     log = Carousel(filewriter),
                     info = Carousel(filewriter),
                     linkages = {("logger", "irc") : ("irc", "inbox"),
                                 ("irc", "outbox") : ("logger", "inbox"),
                                 ("logger", "log_next") : ("log", "next"),
                                 ("logger", "outbox") : ("log", "inbox"),
                                 ("logger", "info_next") : ("info", "next"),
                                 ("logger", "system") : ("info", "inbox"),
                                }
                     ) 
    

config_files = ["/usr/local/etc/Kamaelia/kamaelia_logger.conf",
                "/usr/local/etc/kamaelia_logger.conf",
                "/etc/Kamaelia/kamaelia_logger.conf",
                "/etc/kamaelia_logger.conf",
                "kamaelia_logger.conf",
                "/usr/local/etc/Kamaelia/kamaelia_logger.conf.dist",
                "/usr/local/etc/kamaelia_logger.conf.dist",
                "/etc/Kamaelia/kamaelia_logger.conf.dist",
                "/etc/kamaelia_logger.conf.dist",
                "kamaelia_logger.conf.dist" ]

def openConfig(config_file):
    f = open(config_file)
    lines = f.readlines()
    f.close()
    return lines

def get_config_lines():
    for config_file in config_files:
        try:
            lines = openConfig(config_file)
        except IOError:
            pass
        else:
            config_used =config_file
            break
    return lines

default_config = {
    "channel" : "#kamaelia-test",
    "name": "kamaeliabot" + str(random.randint(1,10000000)),
    "pwd" : None,
    "logdir": "",
}

def parse_config(lines):
   config = copy.deepcopy(default_config)
   for line in lines:
       if line[0] == "#":
           continue
       parts = line.split()
       print parts
       if len(parts) == 3:
          if parts[1] == "=":
             key = parts[0]
             value = parts[2]
             config[key] = value
   return config


__kamaelia_components__ = (BasicLogger, )
__kamaelia_prefabs__ = (Logger, )

if __name__ == '__main__':
    import sys

    raw_config = get_config_lines()
    C = parse_config(raw_config)
#    print config
#    C = default_config
    channel = "#kamaelia-test"
    Name = "kamaeliabot"
    pwd = None
    logdir = "."

    if len(sys.argv) > 1: channel = sys.argv[1]
    if len(sys.argv) > 2: Name = sys.argv[2]
    if len(sys.argv) > 3: pwd = sys.argv[3]

    from Kamaelia.Internet.TCPClient import TCPClient
    from Kamaelia.Util.Introspector import Introspector
    from Kamaelia.Chassis.Pipeline import Pipeline
    Pipeline( Introspector(), TCPClient("127.0.0.1",1501) ).activate()
    print "Logging %s as %s" % (C["channel"], C["name"],)
    Logger(C["channel"],
           name=C["name"],
	   host=C["host"],
           password=C["pwd"],
           logdir=C["logdir"],
           formatter=(lambda data: Kamaelia.Apps.IRCLogger.Support.HTMLOutformat(data)),
           filewriter = Kamaelia.Apps.IRCLogger.Support.LoggerWriter,
           ).run()
